// library.h, v2

#include <string>
#include <map>
#include <iostream>

typedef void(*policyFunction)();
typedef std::map<std::string, 
                 policyFunction> policyMap_t;

// singleton class to manage policies
class Policies {
public:
    void registerPolicy(const std::string id,
                       const policyFunction func);
    void policy(const std::string id) const;

   static Policies& policyMap();
private:
   Policies() {}
   policyMap_t typeMap_;
};

// member function definitions
void
Policies::registerPolicy(const std::string id,
                   const policyFunction func)
{
   if (id != "" && func !=0) {
      typeMap_[id] = func;
   }
}

void
Policies::policy(const std::string id) const
{
   policyMap_t::const_iterator iter = typeMap_.find(id);
   if (iter != typeMap_.end()) {
      (*iter).second();
   }
   else {
       std::cout << "Performing Default Policy." << std::endl << std::endl;
   }
}
    
Policies& 
Policies::policyMap() 
{
   static Policies typeMap;
   return typeMap;
}    

// namespace scope function for convenience
Policies&
policyMap()
{
   return Policies::policyMap();
}
