// library.h, v4

#include <map>
#include <string>
#include <iostream>

typedef void(*policyFunction)();
typedef std::map<const void *,
                 policyFunction> policyMap_t;

// singleton class to manage policies
class Policies {
public:
   void   registerInstance(const void * const v, 
                           const policyFunction func);
   policyFunction isRegistered(const void * const v) const;
   void   policy(const void * const v) const;
   static Policies& instancePolicyMap();
private:
   Policies() {}
   policyMap_t instanceMap_;
};

// member function definitions
void
Policies::registerInstance(const void * const v, 
                           const policyFunction func)
{
   if(v != 0 && func != 0) {
      instanceMap_[v] = func;
   }
}

policyFunction
Policies::isRegistered(const void * const v) const 
{
   policyMap_t::const_iterator iter = instanceMap_.find(v);
   if (iter != instanceMap_.end()) {
      return (*iter).second;
   }
   else {
      return 0;
   }
}

void
Policies::policy(const void * const v) const
{
   if (policyFunction pFunc = isRegistered(v)) {
      pFunc();
   } 
}

Policies&
Policies::instancePolicyMap()
{
   static Policies typeMap;
   return typeMap;
}

// namespace scope function for convenience
Policies& 
instancePolicyMap() 
{
   return Policies::instancePolicyMap();
}    


template<class Container>
void policy(const Container * const c) 
{
   std::cout << "Performing Compile Time Default Policy" 
             << std::endl;
   std::string obj;
   if (c) 
      obj = c->instanceName;
   else
      obj = "unnamed";
   std::cout << "    on Object: " << obj << std::endl << std::endl;
}    
